// $Id: Util.java,v 1.3 2002/06/03 01:51:14 hchen Exp $

import java.io.*;

/**
 * Common tools
 */
public class Util
{
  /**
   * If in debug mode, print a message at stderr
   */
  public static void println(String message, String filename, int lineno)
  {
    if (getDebug())
    {
      System.err.print(message);
      if (filename != null)
	System.err.print(" in file " + filename);
      if (lineno >= 0)
	System.err.print(" at line " + lineno);
      System.err.println();
    }
  }
  
  public static void println(String message)
  {
    if (getDebug())
    {
      System.err.println(message);
    }
  }

  public static void print(String message)
  {
    if (getDebug())
    {
      System.err.print(message);
    }
  }
  
  /**
   * Print a warning message prepended with the class name and method name
   * of the call site
   */
  public static void warn(String message, String filename, int lineno)
  {
    Throwable t = new Throwable();
    warn0(t, message, filename, lineno);
  }

  public static void warn(String message)
  {
    Throwable t = new Throwable();
    warn0(t, message, null, -1);
  }
  
  private static void warn0(Throwable throwable, String message,
			    String filename, int lineno)
  {
    StringBuffer sb;
    int i, begin, length;
    String name;
    boolean tab;

    // Get class and method names at the call site by parsing
    // Throwable messages
    throwable.printStackTrace(pw);
    sb = sw.getBuffer();
    length = sb.length();
    tab = false;
    for (i = 0; i < length; i++)
      if (sb.charAt(i) == '\t')
      {
	if (tab)
	  break;
	else
	  tab = true;
      }

    begin = i + 4;
    for (i = begin; i < length; i++)
      if (sb.charAt(i) == '\n')
	break;
    if (i >= length)
      throw new RuntimeException("Unexpected Throwable format");

    name = sb.substring(begin, i);
    System.err.print("Warning: " + name + ": " + message);
    if (filename != null)
      System.err.print(" in file " + filename);
    if (lineno >= 0)
      System.err.print(" at line " + lineno);
    System.err.println();

    sb.setLength(0);
  }

  /**
   * Throws a RuntimeException
   */
  public static void die(String message, String filename, int lineno)
  {
    String str;

    str = message;
    if (filename != null)
      str += " in file " + filename;
    if (lineno >= 0)
      str += " at line " + lineno;

    throw new RuntimeException(str);
  }

  public static void die(String message)
  {
    die(message, null, -1);
  }

  public static boolean getDebug()
  {
    return isDebug;
  }

  public static void setDebug(boolean isDebug)
  {
    Util.isDebug = isDebug;
  }

  public static void printVersion()
  {
    System.out.println(version);
  }

  static StringWriter sw;
  static PrintWriter pw;
  static boolean isDebug;
  static String version;

  static
  {
    sw = new StringWriter();
    pw = new PrintWriter(sw);
    isDebug = false;
    version =
      "MOPS version 0.9.0: MOdelchecking Programs for Security violations";
  }
}
