// $Id: idemulate.c,v 1.4 2002/03/26 02:06:29 hchen Exp $

#undef _POSIX_SOURCE
#include <sys/capability.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

// There prototypes should have been in one of the above header files
int setresuid(uid_t ruid, uid_t euid, uid_t suid);
int getresuid(uid_t *ruid, uid_t *euid, uid_t *suid);
int setresgid(uid_t ruid, uid_t euid, uid_t suid);
int getresgid(uid_t *ruid, uid_t *euid, uid_t *suid);
extern int capset(cap_user_header_t header, cap_user_data_t data);
extern int capget(cap_user_header_t header, const cap_user_data_t data);

// These macros should have been in the header files
#define CAP_TO_MASK(x) (1 << (x))
#define cap_raise(c, flag)   (c |=  CAP_TO_MASK(flag))
#define cap_lower(c, flag)   (c &= ~CAP_TO_MASK(flag))
#define cap_raised(c, flag)  (c & CAP_TO_MASK(flag))

typedef unsigned char bool;
const int true = 1, false = 0;

typedef struct
{
  int uid, euid, suid, gid, egid, sgid, function, param0, param1, param2;
  int cap_effective, cap_permitted;
} Packet;

// Values that each of ruid, euid, and suid will enumerate on
const int uids[] = { 0, 1 };
// Strings that represent the value of uids in the generated FSA
char * const uidNames[] = {"0", "1"};
const int uidSize = sizeof(uids) / sizeof(uids[0]);

// Values that each of rgid, egid, and sgid will enumerate on
const int gids[] = {};
// Srings that represent the value of gids in the generated FSA
char * const gidNames[] = {};
const int gidSize = sizeof(gids) / sizeof(gids[0]);

// Names of global variables that records uid values
char * const uidVarNames[] = {"saved_euid"}; //, "original_real_uid"};
const int uidVarSize = sizeof(uidVarNames) / sizeof(uidVarNames[0]);

// Names of global variables that records gid values
char * const gidVarNames[] = {}; //{"g1", "g2"};
const int gidVarSize = sizeof(gidVarNames) / sizeof(gidVarNames[0]);

/*
char *disallowedCalls[] =
{
  "capset", "capsetp", "cap_set_proc", "prctl"
};

int disallowedCallsSize = sizeof(disallowedCalls) / sizeof(disallowedCalls[0]);
*/

enum { SETUID, SETEUID, SETREUID, SETRESUID,
       SETGID, SETEGID, SETREGID, SETRESGID, SETIDSIZE
     };

char *syscallNames[SETIDSIZE] =
{ 
  "setuid", "seteuid", "setreuid", "setresuid",
  "setgid", "setegid", "setregid", "setresgid"
};

const int
  ERR_SYSCALL = -1, // set?id() call failed, resulting in error state
  ERR_EOTHER = -2, // EOTHER state
  ERR_IGNORE = -3,  // cannot capset() because of invalid combination of
                    // permitted/effective capabilities
  ERR_DIE = -4;     // program should die immediately because of fatal error

// this value must be negative so as to be distinguished from normal uids
const int INVALID_PARAMETER = -999; 

// pipe fd
int fd[2];

/*
  Sets the user and group ids and capabilities of the process, makes a
  set?id system call, and examines the results

  Returns false on unexpected errors, true otherwise.
*/
bool callback(Packet *packet)
{
  int ret, uid, euid, suid, gid, egid, sgid;
  struct __user_cap_header_struct header;
  struct __user_cap_data_struct data;

  // Set gids.  They must be set first, because setuid() may
  // clear CAP_SETGID bit in the effective capability
  if (packet->gid != INVALID_PARAMETER)
  {
    if (setresgid(packet->gid, packet->egid, packet->sgid) < 0)
    {
      fprintf(stderr,
	"uid=%d euid=%d suid=%d gid=%d egid=%d sgid=%d ",
	 packet->uid, packet->euid, packet->suid,
	 packet->gid, packet->egid, packet->sgid);
      perror("setresgid()");
      packet->uid = ERR_IGNORE;
      return true;
    }
    if (getresgid((uid_t*)&gid, (uid_t*)&egid, (uid_t*)&sgid) < 0)
    {
      perror("getresgid(): ");
      return false;
    }
    if (gid != packet->gid || egid != packet->egid || sgid != packet->sgid)
    {
      fprintf(stderr, "Gids not equal\n");
      return false;
    }
  }

  // Set uids
  if (setresuid(packet->uid, packet->euid, packet->suid) < 0)
  {
    perror("setresuid(): ");
    return false;
  }
  if (getresuid((uid_t*)&uid, (uid_t*)&euid, (uid_t*)&suid) < 0)
  {
    perror("getresuid(): ");
    return false;
  }
  if (uid != packet->uid || euid != packet->euid || suid != packet->suid)
  {
    fprintf(stderr, "uids not equal\n");
    exit(1);
  }

  // Set capabilities
  if (packet->cap_effective != INVALID_PARAMETER)
  {
    header.version = _LINUX_CAPABILITY_VERSION;
    header.pid = 0;
    if (capget(&header, &data) != 0)
    {
      perror("capget(): ");
      return false;
    }
    if (packet->cap_effective)
    {
      cap_raise(data.effective, CAP_SETUID);
      cap_raise(data.effective, CAP_SETGID);
    }
    else
    {
      cap_lower(data.effective, CAP_SETUID);
      cap_lower(data.effective, CAP_SETGID);
    }
    if (packet->cap_permitted)
    {
      cap_raise(data.permitted, CAP_SETUID);
      cap_raise(data.permitted, CAP_SETGID);
    }
    else
    {
      cap_lower(data.permitted, CAP_SETUID);
      cap_lower(data.permitted, CAP_SETGID);
      cap_lower(data.inheritable, CAP_SETUID);
      cap_lower(data.inheritable, CAP_SETGID);
    }

    if (capset(&header, &data) != 0)
    {
      fprintf(stderr,
	    "uid=%d euid=%d suid=%d eff=%x per=%x inh=%x ",
	      packet->uid, packet->euid, packet->suid, data.effective,
	      data.permitted, data.inheritable);
      perror("capset()");
      packet->uid = ERR_IGNORE;
      return true;
    }
  }
  
  switch(packet->function)
  {
    case SETUID:
      ret = setuid(packet->param0);
      break;

    case SETEUID:
      ret = seteuid(packet->param0);
      break;

    case SETREUID:
      ret = setreuid(packet->param0, packet->param1);
      break;

    case SETRESUID:
      ret = setresuid(packet->param0, packet->param1, packet->param2);
      break;

    case SETGID:
      ret = setgid(packet->param0);
      break;

    case SETEGID:
      ret = setegid(packet->param0);
      break;

    case SETREGID:
      ret = setregid(packet->param0, packet->param1);
      break;

    case SETRESGID:
      ret = setresgid(packet->param0, packet->param1, packet->param2);
      break;

    default:
      fprintf(stderr, "Unknown function\n");
      return false;
  }
  
  if (ret == 0)
  {
    if (getresuid((uid_t*)&packet->uid, (uid_t*)&packet->euid, (uid_t*)&packet->suid) < 0)
    {
      perror("getresuid(): ");
      return false;
    }
    // If we care about capabilities
    if (packet->cap_effective != INVALID_PARAMETER)
    {
      if (capget(&header, &data) != 0)
      {
	perror("capget(): ");
	return false;
      }
      // Assume CAP_SETUID & CAP_SETGID bits are always set or cleared together
      packet->cap_permitted = cap_raised(data.permitted, CAP_SETUID) ? true : false;
      packet->cap_effective = cap_raised(data.effective, CAP_SETUID) ? true : false;
    }
    // If we care about gids
    if (packet->gid != INVALID_PARAMETER)
    {
      if (getresgid((uid_t*)&packet->gid, (uid_t*)&packet->egid,
		    (uid_t*)&packet->sgid) < 0)
      {
	perror("getresgid(): ");
	return false;
      }
    }
  }
  else
  {
    packet->uid = ERR_SYSCALL;
    packet->euid = errno;
  }

  return true;
}

// In structure Packet:
// uid, euid, suid: index to uids
// gid, egid, sgid: index to gids
// cap_effective, cap_permitted, param0, param1, param2: values
void test(Packet *inPacket, Packet *outPacket)
{
  Packet packet;
  char *ptr;
  int i, count, ret;

  packet.function = inPacket->function;
  packet.uid = uids[inPacket->uid];
  packet.euid = uids[inPacket->euid];
  packet.suid = uids[inPacket->suid];

  if (inPacket->gid == INVALID_PARAMETER)
  {
    packet.gid = INVALID_PARAMETER;
  }
  else
  {
    packet.gid = gids[inPacket->gid];
    packet.egid = gids[inPacket->egid];
    packet.sgid = gids[inPacket->sgid];
  }

  if (inPacket->cap_effective == INVALID_PARAMETER)
    packet.cap_effective = INVALID_PARAMETER;
  else
  {
    packet.cap_effective = inPacket->cap_effective;
    packet.cap_permitted = inPacket->cap_permitted;
  }
  
  packet.param0 = inPacket->param0;
  packet.param1 = inPacket->param1;
  packet.param2 = inPacket->param2;

  // Warning: You must flush the stdout before forking the child.
  // It took me an evening to figure out this bug!
  fflush(stdout);
  // About to fork the child.  Child and parent communicate through a pipe
  ret = fork();
  if (ret < 0)
  {
    perror("fork()");
    exit(1);
  }
  else if (ret == 0)
  {
    // This is the child
    if (close(fd[0]) < 0)
    {
      perror("close()");
    }
    if (!callback(&packet))
    {
      packet.uid = ERR_DIE;
    }
    if (write(fd[1], &packet, sizeof(packet)) < sizeof(packet))
    {
      perror("write()");
    }
    exit(0);
  }

  // This is the parent
  // parent must call wait() to remove the zombie of the child.
  if (wait(NULL) <= 0)
  {
    perror("wait()");
    exit(1);
  }

  /*
  if (read(fd[0], &packet, sizeof(packet)) < sizeof(packet))
  {
    perror("read()");
    exit(1);
  }
  */
  // Make sure to read in all bytes of structure Packet from the pipe
  count = 0;
  ptr = (char*)&packet;
  while (count < sizeof(packet))
  {
    if ((i = read(fd[0], ptr, sizeof(packet) - count)) < 0)
    {
      perror("read()");
      exit(1);
    }
    count += i;
    ptr += i;
  }

  if (packet.uid == ERR_DIE)
  {
    fprintf(stderr, "Fatal error reported by the child\n");
    exit(1);
  }
  if (packet.uid >= 0)
  {
    // Convert the value of user ids into the indices to the array uids[].
    outPacket->uid = -1;
    outPacket->euid = -1;
    outPacket->suid = -1;
    for (i = 0; i < uidSize; i++)
    {
      if (uids[i] == packet.uid)
	outPacket->uid = i;
      if (uids[i] == packet.euid)
	outPacket->euid = i;
      if (uids[i] == packet.suid)
	outPacket->suid = i;
    }
    if (outPacket->uid == -1 || outPacket->euid == -1 || outPacket->suid == -1)
    {
      fprintf(stderr, "Cannot match uid or euid: uid=%d euid=%d suid=%d\n",
	      packet.uid, packet.euid, packet.suid);
      exit(1);
    }
    if (inPacket->cap_effective == INVALID_PARAMETER)
      outPacket->cap_effective = INVALID_PARAMETER;
    else
    {
      outPacket->cap_effective = packet.cap_effective;
      outPacket->cap_permitted = packet.cap_permitted;
    }

    if (inPacket->gid == INVALID_PARAMETER)
      outPacket->gid = INVALID_PARAMETER;
    else
    {
      // Convert the value of user ids into the indices to the array uids[].
      outPacket->gid = -1;
      outPacket->egid = -1;
      outPacket->sgid = -1;
      for (i = 0; i < gidSize; i++)
      {
	if (gids[i] == packet.gid)
	  outPacket->gid = i;
	if (gids[i] == packet.egid)
	  outPacket->egid = i;
	if (gids[i] == packet.sgid)
	  outPacket->sgid = i;
      }
      if (outPacket->gid == -1 || outPacket->egid == -1 || outPacket->sgid == -1)
      {
	fprintf(stderr, "Cannot match gid or egid: param0=%d param1=%d param2=%d  output: gid=%d egid=%d sgid=%d\n",
		inPacket->param0, inPacket->param1, inPacket->param2,
		packet.gid, packet.egid, packet.sgid);
	exit(1);
      }
    }
  }
  else
  // The set?id system call fails.  Relay the error code to the caller.
  {
    outPacket->uid = packet.uid;
    outPacket->euid = packet.euid;
  }
}

void printTransition(Packet *inPacket, int inVars[],
		     Packet *outPacket, int outVars[],
		     const char *ast)
{
  int i;

  if (outPacket->uid == ERR_IGNORE)
  {
    return;
  }
  
  printf("t \"R=%s,E=%s,S=%s", uidNames[inPacket->uid],
         uidNames[inPacket->euid], uidNames[inPacket->suid]);
  if (inPacket->cap_effective != INVALID_PARAMETER)
    printf(",CE=%s,CP=%s",
         (inPacket->cap_effective ? "1" : "0"),
	 (inPacket->cap_permitted ? "1" : "0"));

  if (inPacket->gid != INVALID_PARAMETER)
  {
    printf(",RG=%s,EG=%s,SG=%s", gidNames[inPacket->gid],
	   gidNames[inPacket->egid], gidNames[inPacket->sgid]);
  }

  for (i = 0; i < uidVarSize; i++)
    printf(",%s=%s", uidVarNames[i], uidNames[inVars[i]]);
  
  if (inPacket->gid != INVALID_PARAMETER)
  {
    for (i = 0; i < gidVarSize; i++)
      printf(",%s=%s", gidVarNames[i], gidNames[inVars[i + uidVarSize]]);
  }

  printf("\" \"");

  if (outPacket->uid >= 0)
  {
    printf("R=%s,E=%s,S=%s", uidNames[outPacket->uid],
	   uidNames[outPacket->euid], uidNames[outPacket->suid]);
    if (outPacket->cap_effective != INVALID_PARAMETER)
      printf(",CE=%s,CP=%s", 
             (outPacket->cap_effective ? "1" : "0"),
	     (outPacket->cap_permitted ? "1" : "0"));

    if (outPacket->gid != INVALID_PARAMETER)
    {
      printf(",G=%s,EG=%s,SG=%s", gidNames[outPacket->gid],
	     gidNames[outPacket->egid], gidNames[outPacket->sgid]);
    }

    for (i = 0; i < uidVarSize; i++)
      printf(",%s=%s", uidVarNames[i], uidNames[outVars[i]]);
    if (outPacket->gid != INVALID_PARAMETER)
    {
      for (i = 0; i < gidVarSize; i++)
        printf(",%s=%s", gidVarNames[i], gidNames[outVars[uidVarSize + i]]);
    }
  }
  else if (outPacket->uid == ERR_SYSCALL)
  {
    switch(outPacket->euid)
    {
      case EPERM:
	printf("EPERM");
	break;

      case EAGAIN:
	printf("EAGAIN");
	break;

      default:
	fprintf(stderr, "Unknown return value %d\n", outPacket->euid);
	exit(1);
    }
  }
  else if (outPacket->uid == ERR_EOTHER)
    printf("EOTHER");
  else
  {
    fprintf(stderr, "Unknown return value %d\n", outPacket->uid);
    exit(1);
  }

  printf("\" %s\n", ast);
}

// This is a very messy function.
void convert(int isGid, int uid, int euid, int suid,
	     int counter[], int selection, int *param, char *label)
{
  int *ids;
  char *name;
  int varSize;
  char **varNames;

  if (isGid)
  {
    ids = (int*)gids;
    name = "gid";
    varSize = gidVarSize;
    varNames = (char**)gidVarNames;
  }
  else
  {
    ids = (int*)uids;
    name = "uid";
    varSize = uidVarSize;
    varNames = (char**)uidVarNames;
  }

  if (selection < varSize)
  {
    *param = ids[counter[selection + (isGid ? uidVarSize : 0)]];
    sprintf(label, "{ identifier \"%s\" }", varNames[selection]);
  }
  else
  {
    switch(selection - varSize)
    {
      case 0:
	*param = ids[uid];
	sprintf(label, "{ function_call { identifier \"get%s\" } }", name);
	break;

      case 1:
	*param = ids[euid];
	sprintf(label, "{ function_call { identifier \"gete%s\" } }", name);
	break;

      case 2:
	*param = -1;
	strcpy(label, "{ unary \"-\" { unsigned_cst \"1\" } }");
	break;

      case 3:
	*param = ids[0];
	// fix me
	sprintf(label, "{ unsigned_cst \"%d\" }", ids[0]);
	break;

      case 4:
	*param = ids[1];
	// fix me
	sprintf(label, "{ unsigned_cst \"%d\" }", ids[1]);
	break;

      default:
	fprintf(stderr, "Unknown selection %d\n", selection - varSize);
	exit(1);
    }
  }
}

void run(bool hasSyscall[])
{
  int *idCounter, *idCounter2;
  char str[1024], label0[1024], label1[1024], label2[1024];
  int i, j, k, mod, uid, euid, suid, gid, egid, sgid, cap_eff, cap_per;
  const int constSize = 5;
  Packet inPacket, outPacket;

  if (pipe(fd) < 0)
  {
    perror("pipe()");
    exit(1);
  }
  idCounter = malloc((uidVarSize + gidVarSize) * sizeof(int));
  idCounter2 = malloc((uidVarSize + gidVarSize) * sizeof(int));
  if (idCounter == NULL || idCounter2 == NULL)
  {
    fprintf(stderr, "Not enough memory\n");
    exit(1);
  }

#ifdef HASUID
  // uid
  for (uid = 0; uid < uidSize; uid++)
  {
    // euid
    for (euid = 0; euid < uidSize; euid++)
    {
      // suid
      for (suid = 0; suid < uidSize; suid++)
      {
#else
	uid = UID;
	euid = EUID;
	suid = SUID;
#endif
        inPacket.uid = uid;
        inPacket.euid = euid;
        inPacket.suid = suid;

#ifdef HASCAP
	// cap_effective
	for (cap_eff = 0; cap_eff <= 1; cap_eff++)
	{
          inPacket.cap_effective = cap_eff;
	  // cap_permitted
	  for (cap_per = 0; cap_per <= 1; cap_per++)
	  {
	    inPacket.cap_permitted = cap_per;
#else
	    cap_eff = INVALID_PARAMETER;
	    inPacket.cap_effective = cap_eff;
#endif	    

#ifdef HASGID
	    for (gid = 0; gid < gidSize; gid++)
	    {
	      inPacket.gid = gid;
	      for (egid = 0; egid < gidSize; egid++)
	      {
		inPacket.egid = egid;
		for (sgid = 0; sgid < gidSize; sgid++)
		{
		  inPacket.sgid = sgid;
#else
		  gid = INVALID_PARAMETER;
		  inPacket.gid = gid;
#endif

		  for (i = 0; i < (uidVarSize + gidVarSize); i++)
		    idCounter[i] = 0;
		  
		  while (1)
		  {
		    // setuid
		    if (hasSyscall[SETUID])
		    {
		      inPacket.function = SETUID;
		      for (i = 0; i < uidVarSize + constSize; i++)
		      {
			convert(false, uid, euid, suid, idCounter,
				i, &inPacket.param0, label0);
			if (inPacket.param0 == -1)
			  continue;
			test(&inPacket, &outPacket);
			sprintf(str,
			  "{ function_call { identifier \"setuid\" } %s } ",
			    label0);
			printTransition(&inPacket, idCounter,
					&outPacket, idCounter,
					str);
		      }
		    }
		    
		    // seteuid
		    if (hasSyscall[SETEUID])
		    {
		      inPacket.function = SETEUID;
		      for (i = 0; i < uidVarSize + constSize; i++)
		      {
			convert(false, uid, euid, suid, idCounter, i,
				&inPacket.param0, label0);
			test(&inPacket, &outPacket);
			sprintf(str,
			 "{ function_call { identifier \"seteuid\" } %s } ",
				label0);
			printTransition(&inPacket, idCounter,
					&outPacket, idCounter, str);
		      }
		    }
		    
		    // setreuid
		    if (hasSyscall[SETREUID])
		    {
		      inPacket.function = SETREUID;
		      for (i = 0; i < uidVarSize + constSize; i++)
			for (j = 0; j < uidVarSize + constSize; j++)
			{
			  convert(false, uid, euid, suid, idCounter, i,
				  &inPacket.param0, label0);
			  convert(false, uid, euid, suid, idCounter, j,
				  &inPacket.param1, label1);
			  test(&inPacket, &outPacket);
			  sprintf(str,
			     "{ function_call { identifier \"setreuid\" } %s %s } ",
				    label0, label1);
			  printTransition(&inPacket, idCounter,
					  &outPacket, idCounter, str);
			}
		    }
		    
		    // setresuid
		    if (hasSyscall[SETRESUID])
		    {
		      inPacket.function = SETRESUID;
		      for (i = 0; i < uidVarSize + constSize; i++)
			for (j = 0; j < uidVarSize + constSize; j++)
			  for (k = 0; k < uidVarSize + constSize; k++)
			  {
			    convert(false, uid, euid, suid, idCounter, i,
				    &inPacket.param0, label0);
			    convert(false, uid, euid, suid, idCounter, j,
				    &inPacket.param1, label1);
			    convert(false, uid, euid, suid, idCounter, k,
				    &inPacket.param2, label2);
			    test(&inPacket, &outPacket);
			    sprintf(str,
			       "{ function_call { identifier \"setresuid\" } %s %s %s } ",
				      label0, label1, label2);
			    printTransition(&inPacket, idCounter,
					    &outPacket, idCounter, str);
			  }
		    }

		    // variable assignments
		    for (i = 0; i < uidVarSize; i++)
		    {
		      for (j = 0; j < (uidVarSize + gidVarSize); j++)
			idCounter2[j] = idCounter[j];
		      idCounter2[i] = uid;
		      sprintf(str, "{ binary \"=\" { identifier \"%s\" } { function_call { identifier \"getuid\" } } }",
			      uidVarNames[i]);
		      printTransition(&inPacket, idCounter,
				      &inPacket, idCounter2, str);
		      idCounter2[i] = euid;
		      sprintf(str, "{ binary \"=\" { identifier \"%s\" } { function_call { identifier \"geteuid\" } } }",
			      uidVarNames[i]);
		      printTransition(&inPacket, idCounter,
				      &inPacket, idCounter2, str);
		    }

		    if (inPacket.gid != INVALID_PARAMETER)
		    {
		      // setgid
		      if (hasSyscall[SETGID])
		      {
			inPacket.function = SETGID;
			for (i = 0; i < gidVarSize + constSize; i++)
			{
			  convert(true, gid, egid, sgid, idCounter,
				  i, &inPacket.param0, label0);
			  if (inPacket.param0 == -1)
			    continue;
			  test(&inPacket, &outPacket);
			  sprintf(str, "{ function_call { identifier \"setgid\" } %s } ",
				  label0);
			  printTransition(&inPacket, idCounter,
					  &outPacket, idCounter,
					  str);
			}
		      }
		      
		      // setegid
		      if (hasSyscall[SETEGID])
		      {
			inPacket.function = SETEGID;
			for (i = 0; i < gidVarSize + constSize; i++)
			{
			  convert(true, gid, egid, sgid, idCounter, i,
				  &inPacket.param0, label0);
			  test(&inPacket, &outPacket);
			  sprintf(str, "{ function_call { identifier \"setegid\" } %s } ",
				  label0);
			  printTransition(&inPacket, idCounter,
					  &outPacket, idCounter, str);
			}
		      }
		      
		      // setregid
		      if (hasSyscall[SETREGID])
		      {
			inPacket.function = SETREGID;
			for (i = 0; i < gidVarSize + constSize; i++)
			  for (j = 0; j < gidVarSize + constSize; j++)
			  {
			    convert(true, gid, egid, sgid, idCounter, i,
				    &inPacket.param0, label0);
			    convert(true, gid, egid, sgid, idCounter, j,
				    &inPacket.param1, label1);
			    test(&inPacket, &outPacket);
			    sprintf(str, "{ function_call { identifier \"setregid\" } %s %s } ",
				    label0, label1);
			    printTransition(&inPacket, idCounter,
					    &outPacket, idCounter, str);
			  }
		      }
		      
		      if (hasSyscall[SETRESGID])
		      {
			// setresgid
			inPacket.function = SETRESGID;
			for (i = 0; i < gidVarSize + constSize; i++)
			  for (j = 0; j < gidVarSize + constSize; j++)
			    for (k = 0; k < gidVarSize + constSize; k++)
			    {
			      convert(true, gid, egid, sgid, idCounter, i,
				      &inPacket.param0, label0);
			      convert(true, gid, egid, sgid, idCounter, j,
				      &inPacket.param1, label1);
			      convert(true, gid, egid, sgid, idCounter, k,
				      &inPacket.param2, label2);
			      test(&inPacket, &outPacket);
			      sprintf(str, "{ function_call { identifier \"setresgid\" } %s %s %s } ",
				      label0, label1, label2);
			      printTransition(&inPacket, idCounter,
					      &outPacket, idCounter, str);
			    }

		      }

		      // variable assignments
		      for (i = 0; i < gidVarSize; i++)
		      {
			for (j = 0; j < (uidVarSize + gidVarSize); j++)
			  idCounter2[j] = idCounter[j];
			idCounter2[uidVarSize + i] = gid;
			sprintf(str, "{ binary \"=\" { identifier \"%s\" } { function_call { identifier \"getgid\" } } }",
			        gidVarNames[i]);
			printTransition(&inPacket, idCounter,
					&inPacket, idCounter2, str);
			idCounter2[uidVarSize + i] = egid;
			sprintf(str, "{ binary \"=\" { identifier \"%s\" } { function_call { identifier \"getegid\" } } }",
			        gidVarNames[i]);
			printTransition(&inPacket, idCounter,
					&inPacket, idCounter2, str);
		      }
		    }

		    // disallowed transitions
		    outPacket.uid = ERR_EOTHER;
		    for (i = 0; i < SETIDSIZE; i++)
		      if (hasSyscall[i])
		      {
			sprintf(str,
			  "{ function_call { identifier \"%s\" } { ellipsis } }",
				syscallNames[i]);
			printTransition(&inPacket, idCounter, &outPacket,
					idCounter, str);
		      }

		    /*
		    for (i = 0; i < disallowedCallsSize; i++)
		    {
		      sprintf(str,
		        "{ function_call { identifier \"%s\" } { ellipsis } }",
			      disallowedCalls[i]);
  		      printTransition(&inPacket, idCounter, &outPacket,
				      idCounter, str);
		    }
		    
		    printTransition(&inPacket, idCounter, &inPacket, idCounter,
				    "{ other }");
		    */

		    // "other" transition
		    printTransition(&inPacket, idCounter, &inPacket, idCounter,
				    "{ other } ");
		    
		    idCounter[0]++;
		    for (i = 0; i < (uidVarSize + gidVarSize); i++)
		    {
		      if (i < uidVarSize)
			mod = uidSize;
		      else
			mod = gidSize;
		      if (idCounter[i] == mod)
		      {
			idCounter[i] = 0;
			if (i < (uidVarSize + gidVarSize - 1))
			  idCounter[i + 1]++;
		      }
		      else
			break;
		    }

		    if (i >= (uidVarSize + gidVarSize))
		      break;
		  }
#ifdef HASGID
		}
	      }
	    }
#endif
#ifdef HASCAP
	  }
	}
#endif
#ifdef HASUID	
      }
    }
  }
#endif
}

int main(int argc, char *argv[])
{
  bool hasSyscall[SETIDSIZE];
  int i;
  char *str;

#ifdef HASUID
  fprintf(stderr, "HASUID ");
#endif
#ifdef HASCAP
  fprintf(stderr, "HASCAP ");
#endif
#ifdef HASGID
  fprintf(stderr, "HASGID ");
#endif
  fprintf(stderr, "\n");
  
  if (argc != 2 || strlen(argv[1]) != SETIDSIZE)
  {
    fprintf(stderr, "Usage: %s [0|1]{8}\n", argv[0]);
    for (i = 0; i < SETIDSIZE; i++)
      fprintf(stderr, "%s ", syscallNames[i]);
    fprintf(stderr, "\n");
    exit(1);
  }

  str = argv[1];
  for (i = 0; i < SETIDSIZE; i++)
    switch(str[i])
    {
      case '0': hasSyscall[i] = false; break;
      case '1': hasSyscall[i] = true; break;
      default:
	fprintf(stderr, "Wrong parameter %s\n", str);
	exit(1);
    }

  fprintf(stderr, "Syscalls: ");
  for (i = 0; i < SETIDSIZE; i++)
    if (hasSyscall[i])
      fprintf(stderr, "%s ", syscallNames[i]);
  fprintf(stderr, "\n");
  
  run(hasSyscall);
}
